/*
  findprimes.c

  Uses the sieve of Eratosthenes to generate a list of primes
*/

// Uncomment the next line to disable assertions
// #define NDEBUG

#include <stdio.h> // printf
#include <stdlib.h> // malloc
#include <assert.h> // assert
#include <math.h> // sqrt and ceil

typedef struct {
  void * next;
  unsigned long int value;
} element;

element * listAppend (element *, unsigned long int);
int listDelete (element *);

int main ()
{
  unsigned long int maxInt = ~ 0x0;
  unsigned long int length = 0;
  unsigned long int loop = 0, limit = 0;

  /* Linked list head and a moving pointer */
  element * list = malloc (sizeof (element));
  element * tail = list;
  element * pointer = NULL;

  maxInt = 285948493;
  
  printf ("Searching from 1 to %lu (%d-bits)...\n", maxInt, sizeof (unsigned long int) * 8);

  printf ("2\n");
  printf ("3\n");

  length = 2;
  list->value = 3; /* Add first prime */

  /* Now we divide all the numbers by all the primes smaller than themselves */
  for (loop = 5; loop <= maxInt; loop += 2)
    {
      pointer = list;

      /* Stop testing at the square root of the digit */
      limit = ceil (sqrt (loop));

      /* Loop until we are past the limit, find a factor or run out of primes */
      while (pointer != NULL &&
	     limit >= pointer->value &&
	     loop % pointer->value != 0)
	{
	  pointer = pointer->next;
	}
      
      /* If we didn't find any factors, it is a prime */
      if (pointer == NULL || limit < pointer->value)
	{
	  ++ length; /* Keep track of the length */
	  tail = listAppend (tail, loop); /* Add the prime to the list */
	  assert (tail != NULL);
	  
 	  printf ("%lu\n", loop);
	}
    }

  listDelete (list);

  return (0);
}

/*
   Appends value to list

   Return values:
   element *	Success
   NULL		Could not allocate memory
*/
element * listAppend (element * list, unsigned long int value)
{
  element * create = NULL;
  
  assert (list != NULL);

  /* Find the end of the list */
  while (list->next != NULL)
    {
      list = list->next;
    }
  
  create = malloc (sizeof (element));

  if (create == NULL)
    {
      return NULL;
    }

  create->next = NULL;
  create->value = value;
  list->next = create;

  return create;
}

/* Deletes a complete list */
int listDelete (element * list)
{
  if (list->next != NULL)
    {
      listDelete (list->next);
      free (list);
    }

  return 0;
}

  

  
